module AsValue::ValueObject

  module InstanceMethods

    def initialize(params = {})
      self.class.send(:attr_accessor, *instance_attributes)

      params.each do |key, value|
        send("#{key}=", value) if instance_attributes.include? key
      end
      
      before_freeze.call(self) unless before_freeze.nil?

      freeze!
    end

    def before_freeze
      self.class.instance_variable_get(:@before_freeze)
    end

    def instance_attributes
      self.class.instance_variable_get(:@instance_attributes)
    end

    def to_h
      Hash[instance_attributes.map { |attribute| [attribute, send(attribute)] }]
    end

    def hash
      to_h.sort.hash
    end

    def ==(other)
      hash == other.hash
    end

    alias_method :eql?, :==

    def freeze!
      instance_attributes.each do |attribute|
        define_singleton_method("#{attribute}=") do |*|
          fail AsValue::ImmutableObjectError,
               "ValueObjects are immutable -- write methods are unavailable."
        end
      end
      self.freeze
    end
  end
  
  def instance_attributes(*attributes)
    @instance_attributes = attributes
  end
  
  def before_freeze(&block)
    @before_freeze = block
  end

  def self.extended(base); base.include(InstanceMethods); end
end
